function [D,u,v]=overrelaxed_sinkhorn_knopp(a,b,C,lambda,theta_max,tolerance,maxIter)
% Compute N dual-Sinkhorn divergences with an overrelaxation version of 
% Cuturi's algortihm
%
%---------------------------
% Required Inputs:
%---------------------------
% a is either
%    - a d1 x 1 column vector in the probability simplex (nonnegative,
%    summing to one). 
%    - a d_1 x N matrix, where each column vector is in the probability simplex
%     
%
% b is a d2 x N matrix of N vectors in the probability simplex
%
% C is a d1 x d2 cost matrix of pairwise distances between bins described 
% in a and bins in the b_1,...b_N histograms.
%
% lambda is the regularization parameter
%
% theta_max\in [1;2) is the maximum value of the overrelaxation parameter
%
%
% tolerance : >0 number to test the stoppingCriterion.
%
% maxIter: maximal number of Sinkhorn fixed point iterations.
%
%---------------------------
% Output
%---------------------------
% D : vector of N dual-sinkhorn divergences
%
% u : d1 x N matrix of left scalings
% v : d2 x N matrix of right scalings
%
% The smoothed optimal transport between (a_i,b_i) can be recovered as
% T_i = diag(u(:,i)) * exp(-C/lambda) * diag(v(:,i));
%
% or, equivalently and substantially faster:
% T_i = bsxfun(@times,v(:,i)',(bsxfun(@times,u(:,i),K)))
%
%
% Relevant paper:
% M. Cuturi,
% Sinkhorn Distances : Lightspeed Computation of Optimal Transport,
% Advances in Neural Information Processing Systems (NIPS) 26, 2013

% A. Thibault, L. Chizat, C. Dossal, and N Papadakis.
%  Overrelaxed sinkhorn-knopp algorithm for regularized optimal transport.
% arXiv preprint arXiv:1711.01851, 2017.

if theta_max<1 || thete_max>=2
    print('The overrelaxation parameter theta_max should be in [1;2)');
    theta_max=max(1,min(1.99999,theta_max));
end
if min(a(:))<=0
    error('The input data a should contain non negative values');
    a=max(1e-16,a);
end

if min(b(:))<=0
    error('The input data b should contain non negative values');
    b=max(1e-16,b);
end

if max(abs(sum(a,1)-1))>1e-16
    error('The input data a does not contain normalized histograms');
    a=a./repmat(sum(a,1), [size(a,1) 1]);
end
if max(abs(sum(b,1)-1))>1e-16
    error('The input data b does not contain normalized histograms');
    b=b./repmat(sum(b,1), [size(b,1) 1]);
end

if size(a,1)~=size(C,1) || size(b,1)~=size(C,2)
   error('The dimension of the cost matric C is not size(a,1) x size(b,1)');
   break;
end

if size(a,2)~=size(b,2)
   error('The dimension of the data does not check size(a,2) = size(b,2)');
   break;
end


K=exp(-C/lambda);


%% Initialization of scaling Factors, N column vectors.
u=ones(size(a,1),size(a,2));
v=ones(size(b,1),size(b,2));




compt=0;
while compt<maxIter,
    compt=compt+1;
    %SK iteration
    u_tilde = a ./ (K * v);
    %overrelaxation
    theta=find_theta();
    u = u.^(1-theta).*(u_tilde).^(theta);
    
     %SK iteration
    v_tilde = b ./ (K' * (u));
    
    %overrelaxation
    theta=find_theta();
    v = v.^(1-theta).*(v_tilde).^(theta);
      
    
    if mod(compt,20)==1 || compt==maxIter,
        
        
        Criterion=norm(sum(abs(v.*(K'*u)-b)),'Inf');
        if Criterion<tolerance || isnan(Criterion),
            break;
        end
        
        
        
        if any(isnan(Criterion))
            error('NaN values have appeared');
        end
    end
    
end

%Deduce the value of the Sinkhorn divergences
D=sum(u.*((C.*K)*v));

   
    
    
end
