function [D,u,v]=overrelaxed_sinkhorn_knopp(a,b,C,epsilon,theta_max,tolerance,maxIter)
% Compute N dual-Sinkhorn divergences with an overrelaxation version of 
% Cuturi's algortihm described in 
% A. Thibault, L. Chizat, C. Dossal, and N Papadakis.
% Overrelaxed sinkhorn-knopp algorithm for regularized optimal transport.
% arXiv preprint arXiv:1711.01851, 2017.


% This function is clearly based on the one of 
% M. Cuturi,
% Sinkhorn Distances : Lightspeed Computation of Optimal Transport,
% Advances in Neural Information Processing Systems (NIPS) 26, 2013
%
%---------------------------
% Required Inputs:
%---------------------------
% a is a d_1 x N matrix, where each column vector is in the probability simplex
% b is a d2 x N matrix of N vectors in the probability simplex
%
% C is a d1 x d2 cost matrix of pairwise distances between bins described 
% in a_i and b_j
%
% epsilon is the regularization parameter
%
% theta_max\in [1;2) is the maximum value of the overrelaxation parameter
%
% tolerance >0 
%
% maxIter: maximal number of iterations.
%
%---------------------------
% Output
%---------------------------
% D : vector of N dual-sinkhorn divergences
%
% u : d1 x N matrix of left scalings
% v : d2 x N matrix of right scalings
%
% The smoothed optimal transport between (a_i,b_i) can be recovered as
% T_i = diag(u(:,i)) * exp(-C/epsilon) * diag(v(:,i));
%
%
% N. Papadakis, 2017.


if theta_max<1 || theta_max>=2
    display('The overrelaxation parameter theta_max should be in [1;2)');
    theta_max=max(1,min(1.99999,theta_max));
end
if min(a(:))<=0
    display('The input data a should contain non negative values');
    a=max(1e-15,a);
end

if min(b(:))<=0
    display('The input data b should contain non negative values');
    b=max(1e-15,b);
end

if max(abs(sum(a,1)-1))>1e-15
    display('The input data a does not contain normalized histograms');
    a=a./repmat(sum(a,1), [size(a,1) 1]);
end
if max(abs(sum(b,1)-1))>1e-15
    display('The input data b does not contain normalized histograms');
    b=b./repmat(sum(b,1), [size(b,1) 1]);
end

if size(a,1)~=size(C,1) || size(b,1)~=size(C,2)
   error('The dimension of the cost matric C is not size(a,1) x size(b,1)');
   return;
end

if size(a,2)~=size(b,2)
   error('The dimension of the data does not check size(a,2) = size(b,2)');
   return;
end

% To prevent from numerical errors
K=max(1e-64,exp(-C/epsilon));


%% Initialization of Left scaling Factors, N column vectors.
u=ones(size(a,1),size(a,2));
v=ones(size(b,1),size(b,2));



delta=0.001;
compt=0;
while compt<maxIter,
    compt=compt+1;
    %SK iteration
    u_tilde = a ./ (K * v);

    %find the theta ensuring the decrease of the Lyapunov function:
    theta=min(theta_max,max(1,find_theta(min(u(:)./u_tilde(:)),theta_max)-delta));
    
    %overrelaxation
    u = u.^(1-theta).*(u_tilde).^(theta);
    
    %SK iteration
    v_tilde = b ./ (K' * (u));
    
    %find the theta ensuring the decrease of the Lyapunov function:
    theta=min(theta_max,max(1,find_theta(min(v(:)./v_tilde(:)),theta_max)-delta));
    %overrelaxation
    v = v.^(1-theta).*(v_tilde).^(theta);
      
    
    if mod(compt,20)==1 || compt==maxIter,
        
        
        Criterion=norm(sum(abs(v.*(K'*u)-b)),'inf');
        if Criterion<tolerance || isnan(Criterion),
            break;
        end
        
        
        
        if any(isnan(Criterion))
            error('NaN values have appeared, the regularization parameter epsilon may be too low');
        end
    end
    
end
 u = a ./ (K * v);
 
%Compute the value of the Sinkhorn divergences
D=sum(u.*((C.*K)*v));

   
    
    
end
