function [D,u,v]=overrelaxed_sinkhorn_knopp(a,b,C,epsilon,theta_max,tolerance,maxIter)
% Compute N dual-Sinkhorn divergences with an overrelaxation version of
% Cuturi's algortihm
%
%---------------------------
% Required Inputs:
%---------------------------
% a is either
%    - a d1 x 1 column vector in the probability simplex (nonnegative,
%    summing to one).
%    - a d_1 x N matrix, where each column vector is in the probability simplex
%
%
% b is a d2 x N matrix of N vectors in the probability simplex
%
% C is a d1 x d2 cost matrix of pairwise distances between bins described
% in a and bins in the b_1,...b_N histograms.
%
% epsilon is the regularization parameter
%
% theta_max\in [1;2) is the maximum value of the overrelaxation parameter
%
%
% tolerance : >0 number to test the stoppingCriterion.
%
% maxIter: maximal number of Sinkhorn fixed point iterations.
%
%---------------------------
% Output
%---------------------------
% D : vector of N dual-sinkhorn divergences
%
% u : d1 x N matrix of left scalings
% v : d2 x N matrix of right scalings
%
% The smoothed optimal transport between (a_i,b_i) can be recovered as
% T_i = diag(u(:,i)) * exp(-C/epsilon) * diag(v(:,i));
%
% or, equivalently and substantially faster:
% T_i = bsxfun(@times,v(:,i)',(bsxfun(@times,u(:,i),K)))
%
%
% Relevant paper:
% M. Cuturi,
% Sinkhorn Distances : Lightspeed Computation of Optimal Transport,
% Advances in Neural Information Processing Systems (NIPS) 26, 2013
%
%Schmitz et al.
% Wasserstein Dictionary Learning: Optimal Transport-based unsupervised
% non-linear dictionary learning,  arXiv preprint arXiv:1708.01955, 2017.
%
% A. Thibault, L. Chizat, C. Dossal, and N Papadakis.
%  Overrelaxed sinkhorn-knopp algorithm for regularized optimal transport.
% arXiv preprint arXiv:1711.01851, 2017.

if theta_max<1 || theta_max>=2
    display('The overrelaxation parameter theta_max should be in [1;2)');
    theta_max=max(1,min(1.99999,theta_max));
end
if min(a(:))<=0
    display('The input data a should contain non negative values');
    a=max(1e-15,a);
end

if min(b(:))<=0
    display('The input data b should contain non negative values');
    b=max(1e-15,b);
end

if max(abs(sum(a,1)-1))>1e-15
    display('The input data a does not contain normalized histograms');
    a=a./repmat(sum(a,1), [size(a,1) 1]);
end
if max(abs(sum(b,1)-1))>1e-15
    display('The input data b does not contain normalized histograms');
    b=b./repmat(sum(b,1), [size(b,1) 1]);
end

if size(a,1)~=size(C,1) || size(b,1)~=size(C,2)
    error('The dimension of the cost matric C is not size(a,1) x size(b,1)');
    return;
end

if size(a,2)~=size(b,2)
    error('The dimension of the data does not check size(a,2) = size(b,2)');
    return;
end




%% Initialization of  scaling Factors, N column vectors.

m=size(a,1);
n=size(b,1);
L=size(a,2);
at=a;
a=reshape(a,[m 1 L]);
bt=b;
b=reshape(b,[1 n L]);

u=zeros(m,1,L);
v=zeros(1,n,L);
C=repmat(C,[1 1 L]);

K=exp(-C/epsilon);

theta=theta_max;

delta=0.001;
compt=0;
while compt<maxIter,
    compt=compt+1;
    
    
    
    vold=v;
    uold=u;
    %log SK iteration
    Ctmp=C-repmat(v, [m 1 1]);
    Ktmp=exp((-Ctmp+repmat(u, [1 n 1]))/epsilon);
    u=u+epsilon*(log(a)-log(sum(Ktmp,2)));
    
    %find the theta ensuring the decrease of the Lyapunov function:
    
    %overrelaxation
    u=u*theta+uold*(1-theta);
    
    %log SK iteration
    Ktmp=exp((-Ctmp+repmat(u, [1 n 1]))/epsilon);
    v=v+epsilon*(log(b)-log(sum(Ktmp,1)));
    
    %find the theta ensuring the decrease of the Lyapunov function:
    
    %overrelaxation
    v=v*theta+vold*(1-theta);
    
    
    
    if mod(compt,20)==1 || compt==maxIter,
        
        u_t=reshape(exp(u/epsilon),[m,L]);
        v_t=reshape(exp(v/epsilon),[n,L]);
        Criterion=norm(sum(abs(v_t.*(K*u_t)-bt)),'inf');
        if Criterion<tolerance || isnan(Criterion),
            break;
        end
        
        
        
        if any(isnan(Criterion))
            error('NaN values have appeared, the regularization parameter epsilon may be too low');
        end
    end
    
end
u=reshape(exp(u/epsilon),[m,L]);
v=reshape(exp(v/epsilon),[n,L]);
%Compute the value of the Sinkhorn divergences
D=sum(u.*((C.*K)*v));




end
