% Test of the Overrelaxed Sinkhorn-Knopp Algorithm
% to compute the Sinkhorn divergence between normalized histograms a and b:

%The Sinkhorn divergence W_\epsilon is a regularization of the optimal transport
%between 2 histograms a and b:

% W_\epsilon(a,b)= \min_T <C,T>-\epsilon <T,\log(T)>,
% under the constraint T1=a and T'1=b;



% If using, please cite

% A. Thibault, L. Chizat, C. Dossal, and N. Papadakis.
% Overrelaxed sinkhorn-knopp algorithm for regularized optimal transport.
% arXiv preprint arXiv:1711.01851, 2017.
%
% M. Cuturi,
% Sinkhorn Distances : Lightspeed Computation of Optimal Transport,
% Advances in Neural Information Processing Systems (NIPS) 26, 2013


% (C) Nicolas Papadakis, 2017


% Grid dimension
M=1000;
x=0:1/(M-1):1;

%cost matrix of size MxM, C(i,j)=(x_i-x_j)^2 
C=toeplitz(x.^2);  

%Regularization parameter for Sinkhorn divergence
epsilon=1e-3;
% Be careful, numerical unstabilities arise for low value of epsilon
% a numerical trick using log stabilizaton is required 
% see Schmitz et al.
% Wasserstein Dictionary Learning: Optimal Transport-based unsupervised
% non-linear dictionary learning,  arXiv preprint arXiv:1708.01955, 2017. 
%


%overrelaxation parameter
theta_max=1.9;

%Tolerance (taken on the \inf norm between the second transport matrix marginal and the
%data b)

tolerance=1e-6;

%Maximum number of iterations
max_iteration=1e4;




%% First test
clear a b
close all
disp(['Test 1) Compute and display the transport matrix between two histograms']);

a = exp(-(x - 0.5).^2 / 0.2^2);
a = a' / sum(a(:));

b = exp(-(x - 0.25).^2 / 0.1^2) + exp(-(x - 0.75).^2 / 0.1^2);
b = b' / sum(b(:));


[D u v]=overrelaxed_sinkhorn_knopp(a,b,C,epsilon,theta_max,tolerance,max_iteration);

%Display 
K=max(1e-64,exp(-C/epsilon));
subplot 221
plot(a);
title('Histogram a')
T=diag(u) * K * diag(v);
subplot 223
imagesc(T');
title('Transport matrix')

subplot 224
plot(b);
title('Histogram b');
camroll(-90)
drawnow;
%% Second Test

% Number of histogram pairs
N=10;
disp(['Test 2) Compute  Sinkhorn divergences between ' num2str(N) ' random histograms of size ' num2str(M)]);
%create random histograms:
%each a(:,i) and b(:,i) are normalized random histograms
a=rand(M,N);
a=a./repmat(sum(a,1), [M 1]); % normalization
a=a./repmat(sum(a,1), [M 1]); % may still be numerical inaccuracies
b=rand(M,N);
b=b./repmat(sum(b,1), [M 1]); % normalization 
b=b./repmat(sum(b,1), [M 1]); % may still be numerical inaccuracies


%compute the Sinkhorn divergence between a(:,i) and b(:,i), i=1:N

[D u v]=overrelaxed_sinkhorn_knopp(a,b,C,epsilon,theta_max,tolerance,max_iteration);
% The smoothed optimal transport between (a(:,i),b(:,i)) can be recovered as
% T_i = diag(u(:,i)) * exp(-C/lambda) * diag(v(:,i));

% display the values of the Sinkhorn divergences
D

